/*
** -----------------------------------------------------------------------------**
** irq_smart.v
**
** making a simgle interrupt that combines frame sync and compressor done
** waiting for the latest of the 2 
**
** Copyright (C) 2008 Elphel, Inc
**
** -----------------------------------------------------------------------------**
**  This file is part of X353
**  X353 is free software - hardware description language (HDL) code.wpage0_inc
** 
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>.
** -----------------------------------------------------------------------------**
**
*/
//     control bits:
//     [1:0] : 3 - enable waiting for frame sync if compressor done comes earlier 
//             2 - disable waiting F.S. (should be disabled for single-frame acquisitions where no frame sync will follow the first frame)
//             1, 0  - don't change F.S. waiting 
//     [3:2] : 3 - wait for DMA FIFO to be transferred to the system before generating interrupt
//             2 - don't wait for the DMA FIFO to be emptied
//             1, 0  - don't change DMA FIFO waiting
//      [15] ; reset requests (mostly fro simulation

module irq_smart        (sclk,               // @negedge
                         wen,                // sync to address and d[0:15]
                         di,                 // [15:0] data in, only [3:0] are currently used
                         frame_sync,         // frame sync, single pulse @ negedge sclk
                         compressor_started, // single pulse @ negedge sclk (this pulse comes before next frame sync if the current frame is being compressed)
                         compressor_done,    // single pulse @ negedge sclk - compressor finished (some data is still in DMA FIFO)
                         fifo_empty,         // DMA FIFO empty (no gaps between compressor_done and !fifo_empty)
                         irq);                // single cycle $ negedge sclk output to be used as IRQ source

    input         sclk;
    input         wen;
    input  [15:0] di;
    input         frame_sync;         // frame sync, single pulse @ negedge sclk
    input         compressor_started; // single pulse @ negedge sclk (this pulse comes before next frame sync if the current frame is being compressed)
    input         compressor_done;    // single pulse @ negedge sclk - compressor finished (some data is still in DMA FIFO)
    input         fifo_empty;         // DMA FIFO empty (no gaps between compressor_done and !fifo_empty)
    output        irq;                // single cycle $ negedge sclk output to be used as IRQ source
	 

    reg           wait_frame_sync;
    reg           wait_fifo;
    wire          done_request;
    reg           compressor_fifo_done;
    wire          compressing;      // from compressor started to compressor done
    wire          waiting_clear;    // frame sync already got, waiting for compressor done (if it was running) to generate request;
//    reg           late_done;
    reg           irq,sirq;
    reg           rst;
FD_1 i_done_request(.C(sclk),.D(!rst && (compressor_done || (done_request && ! compressor_fifo_done)) ), .Q(done_request));
FD_1 i_compressing (.C(sclk),.D(!rst && (compressor_started || (compressing && ! compressor_fifo_done)) ), .Q(compressing));
FD_1 i_waiting_clear (.C(sclk),.D(!rst && (frame_sync || (waiting_clear && ! sirq))), .Q(waiting_clear));
    always @ (negedge sclk) begin
     if (wen & di[1]) wait_frame_sync <= di[0];
     if (wen & di[3]) wait_fifo       <= di[2];
     rst <=wen & di[15];
     compressor_fifo_done <= done_request && (!wait_fifo || fifo_empty);
     sirq <= waiting_clear && (!compressing || frame_sync) ; // includes "missing", if it was waiting till the next frame sync (compressor got stuck? or just stopped by software?)
     irq <= wait_frame_sync? sirq : compressor_fifo_done;
    end
endmodule
