/*
** -----------------------------------------------------------------------------**
** sdram_phase.v
**
** I/O pads related circuitry
**
** Copyright (C) 2002 Elphel, Inc
**
** -----------------------------------------------------------------------------**
**  This file is part of X333
**  X333 is free software - hardware description language (HDL) code.
** 
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>.
** -----------------------------------------------------------------------------**
**
*/

// assumed CL=2.5
module sensor_phase (pre_wcmd,   // decoded address - enables wclk
                     wd,         // CPU write data [3:0]
                                 //       0 - nop, just reset status data
                                 //       1 - increase phase shift
                                 //       2 - decrease phase shift
                                 //       3 - reset phase shift to default (preprogrammed in FPGA configuration)
                                 //       c - reset phase90
                                 //       4 - incr pahse90
                                 //       8 - decrease phase90
                                 //
                     sensordat_pherr, // phase error detected on PXD[2] (should be different)
                                 
                     ph_err,     // [1:0] 0 - no data (SDRAM reads) since last change (wclk*wcmd)
                                 //       1 - clock is too late
                                 //       2 - clock is too early
                                 //       3 - OK (some measurements show too late, some - too early)
                     sclk0,      // global clock, (negedge - commands)
                     pclk,       // global pixel clock (posedge) - from CLK1
                     pixclko,    // pixel clock out (phase shifted from global pclk
//                     pixclki,    // pixel clock input (from some senosors (bpf))
                     dcm_done,
                     locked,     // dcm locked
                     status,     // dcm status (bit 1 - dcm clkin stopped)
                     pclk2x      // pclk multiplied by 2

                     );
   input         pre_wcmd;
   input  [ 3:0] wd;
   input  [ 1:0] sensordat_pherr;
   output [ 1:0] ph_err;
   input         sclk0, pclk;
   output        pixclko;
//   input         pixclki;
   output        dcm_done, locked;
   output [ 7:0] status;
   output        pclk2x;
   
//	reg           wcmd;
	wire          wcmd = pre_wcmd;
   reg    [ 1:0] phase90sel;
   reg       dcm_rst;
//   reg       dcm_rst_cmd;
   wire      dcm_rst_cmd;
   reg [2:0] dcm_drst;
   reg       dcm_en;
   reg       dcm_incdec; 
   wire      pclk2x, pclk2xi;
   wire       pixclk0,pixclk90,pixclk180,pixclk270;
   wire [7:0] status;
   wire       locked;
   reg        dcm_done;
   wire       dcm_done_dcm; // single-cycle
   reg  [2:0] dcm_reset_done;

   reg  [7:0] err_cntr; // up/down counter to compare number of errors early/late
   reg        err; // >0 errors
   reg  [1:0] ph_err;
   
BUFG      i_pclk2x  (.I(pclk2xi), .O(pclk2x));
   
//assign ph_err[1:0]= {(wasearly_rise || wasearly_fall), (waslate_rise || waslate_fall)};
assign pixclko= phase90sel[1]? (phase90sel[0]?pixclk270:pixclk180): (phase90sel[0]?pixclk90:pixclk0);
   
   
// generate control pulses from CPU command

  FD_1 i_dcm_rst_cmd(.Q(dcm_rst_cmd), .D((wcmd && (wd[1:0] == 2'b11)) || (dcm_rst_cmd && !dcm_drst[2])), .C(sclk0)) ;


  always @ (negedge sclk0) begin
//    wcmd <= pre_wcmd;
//    dcm_rst_cmd <= (wcmd && (wd[1:0] == 2'b11)) || (dcm_rst_cmd && !dcm_drst[2]) ;
    dcm_reset_done[2:0] <= {dcm_reset_done[1] & ~dcm_reset_done[0], dcm_reset_done[0], dcm_rst}; // sync to sclk0 end of dcm reset
    
    dcm_en     <= wcmd && (wd[1]!=wd[0]);
    dcm_incdec <= wcmd && wd[0]; 
    if      (wcmd && wd[2] && wd[3]) phase90sel[1:0] <= 2'h0;
    else if (wcmd && wd[2]) phase90sel[1:0] <= phase90sel[1:0] +1;
    else if (wcmd && wd[3]) phase90sel[1:0] <= phase90sel[1:0] -1;
  end

  always @ (posedge pclk) begin
    dcm_drst[2:0] <= dcm_drst[2]? 3'b0:{dcm_drst[1], dcm_drst[0], dcm_rst_cmd};
    dcm_rst    <= dcm_drst[0]  || dcm_drst[1]   || dcm_drst[2] ;
    
  end

// DCM control outputs (use automatic adjustment later?) 
//   assign  dcm_clk=!sclk0; // DCM is triggered by posedge

//sensordat_pherr
//   reg        early_err, late_err;
  always @ (posedge pclk  or posedge wcmd)
    if (wcmd) begin
      err <= 1'b0;
      err_cntr[7:0] <= 8'b0;
    end else begin
// saturate at 0x7f/0x80    
      if      (sensordat_pherr[0] && !sensordat_pherr[1] && (err_cntr[7:0]!=8'h7f)) err_cntr[7:0] <= err_cntr[7:0]+1;
      else if (sensordat_pherr[1] && !sensordat_pherr[0] && (err_cntr[7:0]!=8'h80)) err_cntr[7:0] <= err_cntr[7:0]-1;
      if (|sensordat_pherr[1:0]) err <= 1'b1;
      ph_err[1:0] <= {err,err_cntr[7]};
    end

// make dcm_done behave as dcm_ready

   always @ (negedge sclk0)
     if (dcm_en || dcm_rst_cmd) dcm_done <=1'b0;
     else if (dcm_done_dcm || dcm_reset_done[2]) dcm_done <=1'b1;


DCM i_dcm3(
    .CLKIN    (pclk),
    .CLKFB    (pixclk0),
    .RST      (dcm_rst),
    .PSEN     (dcm_en),
    .PSINCDEC (dcm_incdec),
    .PSCLK    (!sclk0),
    .DSSEN    (1'b0),
    .CLK0     (pixclk0),
    .CLK90    (pixclk90),
    .CLK180   (pixclk180),
    .CLK270   (pixclk270),
    .CLKDV    (),
    .CLK2X    (),
    .CLK2X180 (),
    .CLKFX    (),
    .CLKFX180 (),
    .STATUS   (status[7:0]),
    .LOCKED   (locked),
    .PSDONE   (dcm_done_dcm));
// synthesis attribute CLK_FEEDBACK of i_dcm3 is "1X"
// synthesis attribute CLKIN_DIVIDE_BY_2 of i_dcm3 is "FALSE"
// synthesis attribute CLKIN_PERIOD of i_dcm3 is 10
// synthesis attribute CLKOUT_PHASE_SHIFT of i_dcm3 is "VARIABLE"
// synthesis attribute DESKEW_ADJUST of i_dcm3 is "SYSTEM_SYNCHRONOUS"
// synthesis attribute DLL_FREQUENCY_MODE of i_dcm3 is "LOW"
// synthesis attribute DUTY_CYCLE_CORRECTION of i_dcm3 is "TRUE"

// put here default phase shift ....

// synthesis attribute PHASE_SHIFT of i_dcm3 is 0
// synthesis translate_off
 defparam i_dcm3.CLK_FEEDBACK="1X";
 defparam i_dcm3.CLKIN_DIVIDE_BY_2="FALSE";
 defparam i_dcm3.CLKIN_PERIOD=10; //96Hz
 defparam i_dcm3.CLKOUT_PHASE_SHIFT="VARIABLE";
 defparam i_dcm3.DESKEW_ADJUST="SYSTEM_SYNCHRONOUS";
 defparam i_dcm3.DLL_FREQUENCY_MODE="LOW";
 defparam i_dcm3.DUTY_CYCLE_CORRECTION="TRUE";
 defparam i_dcm3.PHASE_SHIFT=0;
// synthesis translate_on

DCM i_dcm4(
    .CLKIN    (pclk),
    .CLKFB    (pclk2x),
    .RST      (dcm_rst),
    .PSEN     (1'b0),
    .PSINCDEC (1'b0),
    .PSCLK    (1'b0),
    .DSSEN    (1'b0),
    .CLK0     (),
    .CLK90    (),
    .CLK180   (),
    .CLK270   (),
    .CLKDV    (),
    .CLK2X    (pclk2xi),
    .CLK2X180 (),
    .CLKFX    (),
    .CLKFX180 (),
    .STATUS   (),
    .LOCKED   (),
    .PSDONE   ());
    
// If needed - add positive PHASE_SHIFT - then posedge pclk2x will be earlier than posedge pclk by   PHASE_SHIFT/256*period(pclk)  
    
// synthesis attribute CLK_FEEDBACK of i_dcm4 is "2X"
// synthesis attribute CLKIN_DIVIDE_BY_2 of i_dcm4 is "FALSE"
// synthesis attribute CLKIN_PERIOD of i_dcm4 is 10
// synthesis attribute CLKOUT_PHASE_SHIFT of i_dcm4 is "FIXED"
// synthesis attribute DESKEW_ADJUST of i_dcm4 is "SYSTEM_SYNCHRONOUS"
// synthesis attribute DLL_FREQUENCY_MODE of i_dcm4 is "LOW"
// synthesis attribute DUTY_CYCLE_CORRECTION of i_dcm4 is "TRUE"

// put here default phase shift ....

// synthesis attribute PHASE_SHIFT of i_dcm4 is 0
// synthesis translate_off
 defparam i_dcm4.CLK_FEEDBACK="2X";
 defparam i_dcm4.CLKIN_DIVIDE_BY_2="FALSE";
 defparam i_dcm4.CLKIN_PERIOD=10;
 defparam i_dcm4.CLKOUT_PHASE_SHIFT="FIXED";
 defparam i_dcm4.DESKEW_ADJUST="SYSTEM_SYNCHRONOUS";
 defparam i_dcm4.DLL_FREQUENCY_MODE="LOW";
 defparam i_dcm4.DUTY_CYCLE_CORRECTION="TRUE";
 defparam i_dcm4.PHASE_SHIFT=0;
// synthesis translate_on




endmodule
