/*****************************************************************************
*!
*! FILE NAME  : mtd-compat.h
*!
*! DESCRIPTION: Compatibility macros and functions etc for linux mtd in
*!              nandboot environment.
*!
*! ---------------------------------------------------------------------------
*! (C) Copyright 2006-2007, Axis Communications AB, LUND, SWEDEN
*!
*! This program is free software; you can redistribute it and/or modify
*! it under the terms of the GNU General Public License as published by
*! the Free Software Foundation; either version 2 of the License, or
*! (at your option) any later version.
*!
*! This program is distributed in the hope that it will be useful,
*! but WITHOUT ANY WARRANTY; without even the implied warranty of
*! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*! GNU General Public License for more details.
*!
*! You should have received a copy of the GNU General Public License
*! along with this program; if not, write to the Free Software
*! Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*!
*!***************************************************************************/

#ifndef _MTDCOMPAT_H
#define _MTDCOMPAT_H

#include <linux/autoconf.h>

#ifndef __ASSEMBLER__

#include "lib.h"

/* local config, we don't want to depend on linux/config.h for this */
#undef CONFIG_MTD_PARTITIONS
#undef CONFIG_MTD_DEBUG
#undef CONFIG_MTD_DEBUG_VERBOSE
#define CONFIG_MTD_DEBUG_VERBOSE 3 /* debug level */
#undef CONFIG_MTD_NAND_VERIFY_WRITE /* avoid conflict with define in config */
#define CONFIG_MTD_NAND_VERIFY_WRITE

/* our MTD config */

#define NAND_BBT_SUPPORT  0
#define NAND_WRITE_SUPPORT 0
#define NAND_ERASE_SUPPORT 0
#define NAND_MULTICHIP_SUPPORT 0
#define NAND_HWECC_SUPPORT 0
#define NAND_KVEC_SUPPORT 0

/* nullify some functions depending on the above macros */

#if (NAND_ERASE_SUPPORT == 0)
#define nand_erase NULL
#endif

#if (NAND_WRITE_SUPPORT == 0)
#define nand_write NULL
#define nand_write_ecc NULL
#define nand_write_oob NULL
#define nand_block_markbad NULL
#define nand_default_block_markbad NULL
#endif

#if (NAND_KVEC_SUPPORT == 0)
#define nand_writev NULL
#define nand_writev_ecc NULL
#endif

#if (NAND_KVEC_SUPPORT == 0)
struct kvec { int x; };
#endif

/* min and max, from kernel.h */
#define min_t(type,x,y) \
	({ type __x = (x); type __y = (y); __x < __y ? __x: __y; })

#define max_t(type,x,y) \
	({ type __x = (x); type __y = (y); __x > __y ? __x: __y; })

#define min(x,y) ({ \
	typeof(x) _x = (x);	\
	typeof(y) _y = (y);	\
	(void) (&_x == &_y);		\
	_x < _y ? _x : _y; })

#define max(x,y) ({ \
	typeof(x) _x = (x);	\
	typeof(y) _y = (y);	\
	(void) (&_x == &_y);		\
	_x > _y ? _x : _y; })

/* conversions - for little endian machines (i.e. CRIS, and ARM as we use it */
/* from byteorder/generic.h -> byteorder/little_endian.h */
#undef cpu_to_le16
#define cpu_to_le16(x) ((__le16) (__u16)(x))
#undef le16_to_cpu
#define le16_to_cpu(x) ((__u16) (__le16)(x))

/* memory allocation */
#define GFP_KERNEL  0
#define kmalloc(size, mode) malloc(size)
#define kfree(ptr) free(ptr)

/* task management */
#define HZ 1000
#define TASK_INTERRUPTIBLE 0
#define TASK_UNINTERRUPTIBLE 1
#define set_current_state(x) do { } while (0)
#define schedule() do { } while (0)
#define cond_resched() do { } while (0)
#define wake_up(x) do { } while (0)

/* list management */
struct list_head { int x; };
#define wait_queue_head_t int
#define DECLARE_WAITQUEUE(x, y)
#define init_waitqueue_head(x) do { } while (0)
#define init_waitqueue_head(x) do { } while (0)
#define add_wait_queue(x, y) do { } while (0)
#define remove_wait_queue(x, y) do { } while (0)

/* spinlocks */
#define spinlock_t int
#define spin_lock_init(x) do { } while (0)
#define spin_lock(x) do { } while (0)
#define spin_unlock(x) do { } while (0)

/* system time */
unsigned long jiffies; /* note: never bumped, see time_before on next line */
#define time_before(x, y) (x | y | 1)

/* interrupts */
#undef local_irq_save
#define local_irq_save(x) do { (void)x; } while (0)
#undef local_irq_restore
#define local_irq_restore(x) do { (void)x; } while (0)

/* io */
#define __iomem
#define ioremap(addr, pgsize) ((void *) (addr))
#define iounmap(addr) do { } while (0)

/* printk */
/* Just print format string, as a ... well, string */
#ifdef DO_PRINTK /* defined per file when needed */
#define printk(fmt, params...)  simple_printk(fmt)
void simple_printk(const char *fmt);
#else
#define printk(fmt, params...) do { } while (0)
#endif
#define KERN_INFO ""
#define KERN_DEBUG ""
#define KERN_NOTICE ""
#define KERN_WARNING ""
#define KERN_ERR ""

/* module stuff */
#define __init
#define __exit
#define module_init(x)
#define module_exit(x)
#define MODULE_LICENSE(x)
#define MODULE_AUTHOR(x)
#define MODULE_DESCRIPTION(x)
#define EXPORT_SYMBOL(x)
#define EXPORT_SYMBOL_GPL(x)

/* misc */
#define touch_softlockup_watchdog() do { } while (0)
#undef THIS_MODULE
#define THIS_MODULE 0
struct notifier_block { int x; };

#undef BUG
#define BUG() error2("BUG in " __FILE__, __LINE__, __FUNCTION__)

#endif /* !__ASSEMBLER__ */

#endif /* _MTDCOMPAT_H */
